/**
 * @file Tile.java
 *
 * @author Christos Choutouridis AEM:8997
 * @email  cchoutou@ece.auth.gr
 */

package net.hoo2.auth.labyrinth;

/**
 * @brief
 *    This class is the representation of the board's tile
 *
 * Tiles are arranged on the board in square shape and they're identified by
 * an ID. This ID is the linear combination of x and y coordinate of the tile.
 */
class Tile {

   /** @name Constructors */
   /** @{ */

   /**
    * The main constructor of the Tile constructed from (row,column)
    *
    * @param row     The row coordinate to place the Tile
    * @param col     The column coordinate to place the Tile
    * @param up      The existence of wall north of the tile
    * @param down    The existence of wall south of the tile
    * @param left    The existence of wall left of the tile
    * @param right   The existence of wall right of the tile
    */
   protected Tile(int row, int col, boolean up, boolean down, boolean left, boolean right) {
      // Boundary checks
      assert (row >= 0 && row< Session.boardSize) : "Row coordinate must be in the range [0, Session.boardSize)";
      assert (col >= 0 && col< Session.boardSize) : "Column coordinate must be in the range [0, Session.boardSize)";

      // Initialization
      this.tileId = Position.toID(row, col);
      this.x =col;
      this.y =row;
      this.up = up;
      this.down = down;
      this.left = left;
      this.right = right;
   }

   /**
    * The main constructor of the Tile constructed from tileId
    *
    * @param id      The tileId to place the Tile
    * @param up      The existence of wall north of the tile
    * @param down    The existence of wall south of the tile
    * @param left    The existence of wall left of the tile
    * @param right   The existence of wall right of the tile
    */
   protected Tile(int id, boolean up, boolean down, boolean left, boolean right) {
      // Boundary check
      assert (id >= 0 && id <= Position.toID(Session.boardSize-1, Session.boardSize-1))
         : "TileId must be in the range of [0, Session.boardSize^2)";
      // Initialization
      this.tileId = id;
      this.x      =Position.toCol(id);
      this.y      =Position.toRow(id);
      this.up     = up;
      this.down   = down;
      this.left   = left;
      this.right  = right;
   }

   /**
    * A deep copy constructor.
    */
   protected Tile (Tile t) {
      this.tileId    = t.tileId;
      this.x         = t.x;
      this.y         = t.y;
      this.up        = t.up;
      this.down      = t.down;
      this.left      = t.left;
      this.right     = t.right;
      // We achieve deep copy as the members are all primitives.
   }
   /** @} */

   /** @name Supply's main application interface */
   /** @{ */

   /**
    * @return the position of the tile as a Position object
    * @see Position
    */
   protected Position position ()   { return new Position (tileId); }

   /**
    * Set the position of the tile from a (row, column) pair
    * @param row  The row coordinate of the tile
    * @param col  The column coordinate of the tile
    * @return the position of the supply as a Position object
    * @note       This function also returns the supplyId to help in chained expressions.
    * @see Position
    */
   protected Position position (int row, int col) {
      // Boundary checks
      assert (row >= 0 && row< Session.boardSize) : "Row coordinate must be in the range [0, Session.boardSize)";
      assert (col >= 0 && col< Session.boardSize) : "Column coordinate must be in the range [0, Session.boardSize)";

      Position p  = new Position (row, col);
      this.x      = p.getCol();  // =col;
      this.y      = p.getRow();  // =row;
      this.tileId = p.getId();
      return p;
   }

   /**
    * Set the position of the tile from a tileId
    * @param tileId  The tileId position
    * @return        The position of the supply as Position object
    * @note          This function also returns the supplyId to help in chained expressions.
    * @see Position
    */
   protected Position position (int tileId) {
      // Boundary check
      assert (tileId >= 0 && tileId <= Position.toID(Session.boardSize-1, Session.boardSize-1))
         : "TileId must be in the range of [0, Session.boardSize^2)";

      Position p  = new Position (tileId);
      this.x      = p.getCol();
      this.y      = p.getRow();
      this.tileId = p.getId();   // =tileId;
      return p;
   }

   /**
    * Set the tile's walls
    * @param up      Request for north wall (winter is coming)
    * @param down    Request for south wall
    * @param left    Request for west wall
    * @param right   Request for east wall
    */
   protected void setWalls (boolean up, boolean down, boolean left, boolean right) {
      assert (
         ((up?1:0) + (down?1:0) + (left?1:0) + (right?1:0)) <=2
      ) : "A tile can have at most 2 walls";

      this.up     = (up)   ? true : this.up;
      this.down   = (down) ? true : this.down;
      this.left   = (left) ? true : this.left;
      this.right  = (right)? true : this.right;
   }

   /**
    * Checks if the tile has wall in the requested direction
    * @param direction  The direction to check
    * @return     True if there is a wall
    */
   protected boolean hasWall (int direction) {
      switch (direction) {
         case Direction.UP:    return up;
         case Direction.RIGHT: return right;
         case Direction.DOWN:  return down;
         case Direction.LEFT:  return left;
      }
      return false;
   }

   /**
    * Checks if the tile has walls and return the number of them
    * @return     The number of walls
    */
   protected int hasWalls () {
      return ((up?1:0) + (down?1:0) + (left?1:0) + (right?1:0));
   }

   protected int hasSupply (Supply[] supplies) {
      for (Supply s: supplies)
         if (s.position().getId() == position().getId())
            return s.getSupplyId();
      return Const.noSupply;
   }
   /** @} */

   /**
    * @name Accessor/Mutator interface
    * @note
    *    Please consider not to use mutator interface. Its the abstraction killer :(
    *    We have added a bit of logic however, in order to make it a bit more safe.
    */
   /** @{ */
   protected int getTileId () { return tileId; }
   protected int getX ()      { return x; }
   protected int getY ()      { return y; }
   protected boolean getUp ()    { return up; }
   protected boolean getDown ()  { return down; }
   protected boolean getLeft ()  { return left; }
   protected boolean getRight () { return right; }

   protected void setTileId(int tileId) {
      assert (tileId >= 0 && tileId <= Position.toID(Session.boardSize-1, Session.boardSize-1))
      : "TileId must be in the range of [0, Session.boardSize^2)";
      this.tileId = tileId;
      this.x = Position.toCol(tileId);
      this.y = Position.toRow(tileId);
   }
   protected void setX(int x) {
      assert (x >= 0 && x< Session.boardSize) : "X coordinate must be in the range [0, Session.boardSize)";
      this.x = x;
      this.tileId = Position.toID(this.x, this.y);
   }
   protected void setY(int y) {
      assert (y >= 0 && y< Session.boardSize) : "Y coordinate must be in the range [0, Session.boardSize)";
      this.y = y;
      this.tileId = Position.toID(this.x, this.y);
   }
   protected void setUp(boolean up)       { this.up = up; }
   protected void setDown(boolean down)   { this.down = down; }
   protected void setRight(boolean right) { this.right = right; }
   protected void setLeft(boolean left)   { this.left = left; }
   /** @} */

   /** @name Class data */
   /** @{ */
   private int tileId;        /**<
                               * The unique identifier of the tile. This is the linear combination of
                               * x and y coordinates of the tile
                               */
   private int x;             /**< The x coordinate(column) of the tile as if the board lies in the 1st quadrant */
   private int y;             /**< The y coordinate(row) of the tile as if the board lies in the 1st quadrant */
   private boolean up;        /**< Indicator of a wall in the north side of the tile */
   private boolean down;      /**< Indicator of a wall in the south side of the tile */
   private boolean left;      /**< Indicator of a wall in the left side of the tile */
   private boolean right;     /**< Indicator of a wall in the right side of the tile */
   /**
    * @warning
    *    We can calculate tileId from (x,y) so having both (x,y) and tile ID is error
    *    prone and not a good practice. Its easy to get them out of sync(code smell).
    *    We implement it just because its in the requirements of the assignment.
    *    @see Supply.supplyTileId
    */
}
